from __future__ import print_function
from functools import partial
import inspect
import os
import colorama
import sys
import slash
from slash.utils.python import get_underlying_func


_COLOR_RESET = colorama.Fore.RESET + colorama.Back.RESET + colorama.Style.RESET_ALL  # pylint: disable=no-member


def _title_style(s):
    # pylint: disable=no-member
    return '{0}{1}{2}{3}'.format(colorama.Fore.WHITE, colorama.Style.BRIGHT, s, _COLOR_RESET)


def _unused_style(s):
    # pylint: disable=no-member
    return '{0}{1}{2}'.format(colorama.Fore.YELLOW, s, _COLOR_RESET)


def _doc_style(s):
    # pylint: disable=no-member
    return '{0}{1}{2}{3}'.format(colorama.Fore.GREEN, colorama.Style.BRIGHT, s, _COLOR_RESET)


def slash_fixtures(args, report_stream=sys.stdout):
    _print = partial(print, file=report_stream)

    path = args[0] if args else '.'
    with slash.Session() as session:
        slash.site.load()
        loader = slash.loader.Loader()
        runnables = loader.get_runnables([path])
        used_fixtures = set()
        for test in runnables:
            used_fixtures.update(test.get_required_fixture_objects())

        for fixture in session.fixture_store:
            if not hasattr(fixture, 'fixture_func'):
                continue

            fixture_func = get_underlying_func(fixture.fixture_func)
            doc = inspect.cleandoc(fixture_func.__doc__) if fixture_func.__doc__ else ''

            unused_string = '' if fixture in used_fixtures else ' (Unused)'

            _print(_title_style('{0}{1}'.format(fixture.info.name, unused_string)))
            if doc:
                for line in (_doc_style(doc)).split('\n'):
                    _print('    {0}'.format(line))

            _print('    Source: {0}:{1}'.format(
                os.path.relpath(inspect.getsourcefile(fixture_func), path),
                inspect.getsourcelines(fixture_func)[1]))
            _print('\n')



import gossip
import slash
import slash.runner
from slash import exception_handling, Session
from slash.loader import Loader

from .utils import TestCase


def test_test_cleanups_happen_before_fixture_cleanups(suite, suite_test):
    fixture = suite.slashconf.add_fixture()
    suite_test.depend_on_fixture(fixture)
    fixture_cleanup = fixture.add_cleanup()
    test_cleanup = suite_test.add_cleanup()
    events = suite.run().events

    assert events[fixture_cleanup].timestamp > events[test_cleanup].timestamp


def test_cleanups(suite, suite_test):

    cleanup = suite_test.add_cleanup()

    summary = suite.run()

    assert summary.events[cleanup]


def test_cleanup_args_kwargs_deprecated():
    with slash.Session() as s:
        slash.add_cleanup(lambda: None, "arg1", arg2=1)
    [w] = s.warnings
    assert 'deprecated' in str(w).lower()


class CleanupsTest(TestCase):

    def setUp(self):
        super(CleanupsTest, self).setUp()
        self._successful_tests = []
        self.addCleanup(gossip.unregister_token, "monitor_test_success")
        slash.hooks.test_success.register(
            self._register_test_success, token="monitor_test_success")

    def _register_test_success(self):
        self._successful_tests.append(slash.context.test)

    def test_cleanups(self):
        class Test(slash.Test):

            def test1(self_):
                self.events.test1()
                slash.add_cleanup(self.events.cleanup,
                                  args=("test1 cleanup 2",))
                slash.add_cleanup(self.events.cleanup,
                                  args=("test1 cleanup 1",))

            def test2(self_):
                self.events.test2()
                slash.add_cleanup(self.events.cleanup, args=("test2 cleanup",))

        with self.forge.any_order():
            with self.forge.ordered():
                self.events.test1()
                self.events.cleanup("test1 cleanup 1")
                self.events.cleanup("test1 cleanup 2")
            with self.forge.ordered():
                self.events.test2()
                self.events.cleanup("test2 cleanup")

        self.forge.replay()
        with Session() as s:
            with s.get_started_context():
                slash.runner.run_tests(Loader().get_runnables(Test))

        assert len(self._successful_tests) == 2

    def test_error_cleanups_and_fail_test(self):
        self._test_error_cleanups(fail_test=True)

    def test_error_cleanups_and_not_fail_test(self):
        self._test_error_cleanups(fail_test=False)

    def _test_error_cleanups(self, fail_test):

        exc_infos = []
        new_exc_handlers = list(exception_handling._EXCEPTION_HANDLERS)
        new_exc_handlers.append(exc_infos.append)
        self.forge.replace_with(
            exception_handling, "_EXCEPTION_HANDLERS", new_exc_handlers)

        class Test(slash.Test):

            def test(self_):
                slash.add_cleanup(self.events.cleanup, args=(1,))
                slash.add_cleanup(self.events.cleanup, args=(2,))
                slash.add_success_only_cleanup(self.events.cleanup, args=(3,))
                if fail_test:
                    raise Exception("!!!")
        if not fail_test:
            self.events.cleanup(3).and_raise(ThirdException())
        self.events.cleanup(2).and_raise(SecondException())
        self.events.cleanup(1).and_raise(FirstException())

        self.forge.replay()

        with Session() as session:
            with session.get_started_context():
                slash.runner.run_tests(Loader().get_runnables(Test))
        self.forge.verify()
        [result] = session.results.iter_test_results()
        errors = result.get_errors()

        assert len(errors) == 3

        self.assertEquals(len(errors), len(exc_infos))
        self.assertEquals(
            [e[0] for e in exc_infos],
            [Exception, SecondException, FirstException] if fail_test else [
                ThirdException, SecondException, FirstException],
        )
        self.assertEquals(self._successful_tests, [])


class FirstException(Exception):
    pass


class SecondException(Exception):
    pass


class ThirdException(Exception):
    pass



from .markers import repeat
from .metadata import Metadata
from ..conf import config


class RunnableTestFactory(object):

    def __init__(self, file_path='', module_name='', factory_name=''):
        super(RunnableTestFactory, self).__init__()
        self.file_path = file_path
        self.module_name = module_name
        self.factory_name = factory_name

    def generate_tests(self, fixture_store):
        """
        Generates :class:`.RunnableTest` instances to run

        Do not override this method directly. Use :func:`.RunnableTestFactory._generate_tests` instead.
        """
        for address_in_factory, test in self._generate_tests(fixture_store):
            assert test.__slash__ is None
            test.__slash__ = Metadata(self, test, address_in_factory)
            yield test

    def _generate_tests(self, fixture_store):
        raise NotImplementedError()  # pragma: no cover

    def _get_num_repetitions(self, func):
        return repeat.get_value(func, 1) * config.root.run.repeat_each




import itertools
import platform
import shutil
import tempfile

import forge
import logbook
from logbook.compat import LoggingHandler

import gossip
import slash
from slash.conf import config
from slash._compat import ExitStack

if platform.python_version() < "2.7":
    import unittest2 as unittest
else:
    import unittest


_logger = logbook.Logger(__name__)

class TestCase(unittest.TestCase):
    def setUp(self):
        super(TestCase, self).setUp()
        self._handler = LoggingHandler()
        self._handler.push_application()
        self.addCleanup(self._handler.pop_application)
        gossip.get_group('slash').set_exception_policy(gossip.RaiseImmediately())
        self.override_config("log.console_level", 10000) # silence console in tests

    def override_config(self, path, value):
        self.addCleanup(config.assign_path, path, config.get_path(path))
        config.assign_path(path, value)

    def get_new_path(self):
        returned = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, returned)
        return returned

    _forge = None
    @property
    def forge(self):
        if self._forge is None:
            self._forge = forge.Forge()
        return self._forge
    _events = None
    @property
    def events(self):
        if self._events is None:
            self._events = self.forge.create_wildcard_mock()
        return self._events
    def tearDown(self):
        if self._forge is not None:
            self._forge.restore_all_replacements()
            self._forge.verify()
        super(TestCase, self).tearDown()

class NullFile(object):
    def write(self, s):
        pass

    def isatty(self):
        return False

    def flush(self):
        pass

class CustomException(Exception):
    @classmethod
    def do_raise(cls):
        raise cls("Custom exception")

def no_op(*args, **kwargs):
    pass

def run_tests_in_session(test_class_path_or_iterator, session=None):
    with ExitStack() as stack:
        if session is None:
            session = slash.Session()
            stack.enter_context(session)
        

        if isinstance(test_class_path_or_iterator, str):
            test_class_path_or_iterator = slash.loader.Loader().iter_paths([test_class_path_or_iterator])
        else:
            test_class_path_or_iterator = slash.loader.Loader().get_runnables(test_class_path_or_iterator)
        with session.get_started_context():
            slash.runner.run_tests(test_class_path_or_iterator)
    for result in session.results.iter_test_results():
        for err in itertools.chain(result.get_errors(), result.get_failures(), result.get_skips()):
            _logger.debug("Unsuccessful result: {0}", err)
    return session
run_tests_in_session.__test__ = False

def run_tests_assert_success(test_class_path_or_iterator, session=None):
    session = run_tests_in_session(test_class_path_or_iterator, session=session)
    assert session.results.is_success(), "Run did not succeed"
    return session

run_tests_assert_success.__test__ = False



class FixtureScopeManager(object):

    def __init__(self, fixture_store):
        super(FixtureScopeManager, self).__init__()
        self._fixture_store = fixture_store
        self._last_module = self._last_test = None

    def begin_test(self, test):
        test_module = test.__slash__.module_name
        if self._last_module is None:
            self._fixture_store.begin_scope('session')
            self._fixture_store.begin_scope('module')
        elif self._last_module != test_module:
            self._fixture_store.begin_scope('module')
        self._last_module = test_module
        self._fixture_store.begin_scope('test')
        self._last_test = test

    def end_test(self, test, next_test):
        assert test == self._last_test
        self._fixture_store.end_scope('test')
        if next_test is None or next_test.__slash__.module_name != self._last_module:
            self._fixture_store.end_scope('module')
        if next_test is None:
            self._fixture_store.end_scope('session')



from slash.frontend.slash_fixtures import slash_fixtures
from slash._compat import StringIO


def test_fixture_cleanup_at_end_of_suite(suite):
    suite.debug_info = False
    f = suite.slashconf.add_fixture()
    path = suite.commit()
    report_stream = StringIO()

    slash_fixtures([path], report_stream)
    assert 'Source' in report_stream.getvalue()



from sentinels import NOTHING

from ..._compat import iteritems, itervalues
from ...exceptions import CyclicFixtureDependency, UnresolvedFixtureStore
from ...utils.python import getargspec
from .fixture import Fixture
from .namespace import Namespace
from .parameters import Parametrization, get_parametrization_fixtures
from .utils import get_scope_by_name, nofixtures
from .variation import VariationFactory


class FixtureStore(object):

    def __init__(self):
        super(FixtureStore, self).__init__()
        self._namespaces = [Namespace(self)]
        self._unresolved_fixture_ids = set()
        self._fixtures_by_fixture_info = {}
        self._fixtures_by_id = {}
        self._values_by_id = {}
        self._cleanups_by_scope = {}
        self._all_needed_parametrization_ids_by_fixture_id = {}

    def call_with_fixtures(self, test_func, namespace, is_method):

        if not nofixtures.is_marked(test_func):
            arg_names = self.get_required_fixture_names(test_func, is_method=is_method)
            kwargs = self.get_fixture_dict(arg_names, namespace)
        else:
            kwargs = {}

        return test_func(**kwargs)  # pylint: disable=star-args

    def get_required_fixture_names(self, test_func, is_method):
        skip_names = set(p.name for p in get_parametrization_fixtures(test_func))
        arg_names = [name for name in getargspec(test_func).args if name not in skip_names]
        if is_method:
            arg_names = arg_names[1:]
        return arg_names

    def get_required_fixture_objects(self, test_func, namespace, is_method):
        names = self.get_required_fixture_names(test_func, is_method=is_method)
        return set(itervalues(self.get_fixture_dict(names, namespace=namespace, get_values=False)))

    def __iter__(self):
        return itervalues(self._fixtures_by_id)

    def push_namespace(self):
        self._namespaces.append(Namespace(self, parent=self._namespaces[-1]))

    def pop_namespace(self):
        return self._namespaces.pop(-1)

    def get_current_namespace(self):
        return self._namespaces[-1]

    def add_cleanup(self, scope, cleanup):
        assert isinstance(scope, int)
        self._cleanups_by_scope.setdefault(scope, []).append(cleanup)

    def get_all_needed_parametrization_ids(self, fixtureobj):
        if self._unresolved_fixture_ids:
            raise UnresolvedFixtureStore()
        if isinstance(fixtureobj, Parametrization):
            return frozenset([fixtureobj.info.id])
        returned = self._all_needed_parametrization_ids_by_fixture_id.get(fixtureobj.info.id)
        if returned is None:
            returned = self._compute_all_needed_parametrization_ids(fixtureobj)
            self._all_needed_parametrization_ids_by_fixture_id[fixtureobj.info.id] = returned
        return returned

    def iter_autouse_fixtures_in_namespace(self, namespace=None):
        if namespace is None:
            namespace = self.get_current_namespace()
        for fixture in namespace.iter_fixtures():
            if fixture.info.autouse:
                yield fixture

    def activate_autouse_fixtures_in_namespace(self, namespace):
        for fixture in self.iter_autouse_fixtures_in_namespace(namespace):
            _ = self.get_fixture_value(fixture)

    def _compute_all_needed_parametrization_ids(self, fixtureobj):
        stack = [fixtureobj.info.id]
        returned = set()
        while stack:
            fixture_id = stack.pop()
            if fixture_id in self._all_needed_parametrization_ids_by_fixture_id:
                returned.update(self._all_needed_parametrization_ids_by_fixture_id[fixture_id])
                continue
            fixture = self._fixtures_by_id[fixture_id]
            if fixture.parametrization_ids:
                returned.update(fixture.parametrization_ids)
            if fixture.fixture_kwargs:
                stack.extend(itervalues(fixture.fixture_kwargs))
        return frozenset(returned)

    def begin_scope(self, scope):
        scope = get_scope_by_name(scope)

    def end_scope(self, scope):
        scope = get_scope_by_name(scope)
        cleanups = self._cleanups_by_scope.get(scope, [])
        for fixture_id, fixture in iteritems(self._fixtures_by_id):
            if fixture.scope <= scope:
                self._values_by_id.pop(fixture_id, None)
        while cleanups:
            cleanups.pop(-1)()

    def ensure_known_parametrization(self, parametrization):
        if parametrization.info.id not in self._fixtures_by_id:
            self._fixtures_by_id[parametrization.info.id] = parametrization

    def add_fixtures_from_dict(self, d):
        for name, thing in iteritems(d):
            fixture_info = getattr(thing, '__slash_fixture__', None)
            if fixture_info is None:
                continue
            self.get_current_namespace().add_name(
                name, self.add_fixture(thing).__slash_fixture__.id)

    def add_fixture(self, fixture_func):
        fixture_info = fixture_func.__slash_fixture__
        existing_fixture = self._fixtures_by_id.get(fixture_info.id)
        if existing_fixture is not None:
            return existing_fixture.fixture_func
        fixture_object = Fixture(self, fixture_func)
        current_namespace = self._namespaces[-1]
        current_namespace.add_name(fixture_info.name, fixture_info.id)
        self.register_fixture_id(fixture_object)
        return fixture_func

    def register_fixture_id(self, f):
        assert f.info.id not in self._fixtures_by_id
        self._fixtures_by_id[f.info.id] = f
        self._unresolved_fixture_ids.add(f.info.id)

    def get_fixture_by_name(self, name):
        return self._namespaces[-1].get_fixture_by_name(name)

    def get_fixture_by_id(self, fixture_id):
        return self._fixtures_by_id[fixture_id]

    def get_fixture_dict(self, required_names, namespace=None, get_values=True, skip_names=frozenset()):
        returned = {}

        if namespace is None:
            namespace = self.get_current_namespace()

        for required_name in required_names:
            if required_name in skip_names:
                continue
            fixture = namespace.get_fixture_by_name(required_name)
            if get_values:
                fixture = self.get_fixture_value(fixture, name=required_name)
            returned[required_name] = fixture
        return returned

    def get_fixture_value(self, fixture, name=None):
        if name is None:
            name = fixture.info.name

        self._fill_fixture_value(name, fixture)
        return self._values_by_id[fixture.info.id]

    def iter_parametrization_variations(self, fixture_ids=(), funcs=(), methods=()):

        variation_factory = VariationFactory(self)
        for fixture_id in fixture_ids:
            variation_factory.add_needed_fixture_id(fixture_id)

        for func in funcs:
            variation_factory.add_needed_fixtures_from_function(func)

        for method in methods:
            variation_factory.add_needed_fixtures_from_method(method)

        return variation_factory.iter_variations()

    def _fill_fixture_value(self, name, fixture):

        fixture_value = self._values_by_id.get(fixture.info.id, NOTHING)
        if fixture_value is _BUSY:
            raise CyclicFixtureDependency(
                'Fixture {0!r} is a part of a dependency cycle!'.format(name))

        if fixture_value is not NOTHING:
            return

        self._values_by_id[fixture.info.id] = _BUSY

        kwargs = {}

        if fixture.fixture_kwargs is None:
            raise UnresolvedFixtureStore('Fixture {0} is unresolved!'.format(name))

        for required_name, fixture_id in iteritems(fixture.fixture_kwargs):
            self._fill_fixture_value(
                required_name, self.get_fixture_by_id(fixture_id))
            kwargs[required_name] = self._values_by_id[fixture_id]

        self._values_by_id[fixture.info.id] = fixture.get_value(kwargs)

    def resolve(self):
        while self._unresolved_fixture_ids:
            fixture = self._fixtures_by_id[self._unresolved_fixture_ids.pop()]
            fixture.resolve(self)


_BUSY = object()



import slash

from .utils import TestCase
from slash import Session
from slash.core.result import Result
from slash.core.result import SessionResults

from .utils import run_tests_assert_success

def test_result_summary(suite):

    suite[2].when_run.fail()
    suite[3].when_run.raise_exception()
    suite[4].when_run.raise_exception()
    suite[5].when_run.skip()

    results = suite.run().session.results

    assert results.get_num_errors() == 2
    assert results.get_num_failures() == 1
    assert results.get_num_skipped() == 1
    assert results.get_num_successful() == len(suite) - 4

def test_has_errors_or_failures(suite):
    suite[2].when_run.fail()
    suite[3].when_run.raise_exception()
    results = suite.run().session.results
    assert not results[0].has_errors_or_failures()
    assert results[2].has_errors_or_failures()
    assert results[3].has_errors_or_failures()


def test_has_skips(suite):
    suite[1].when_run.skip()
    results = suite.run().session.results
    assert not results[0].has_skips()
    assert results[1].has_skips()


def test_result_data_is_unique():

    class SampleTest(slash.Test):

        def test_1(self):
            pass

        def test_2(self):
            pass

    session = run_tests_assert_success(SampleTest)
    [result1, result2] = session.results
    assert result1.data is not result2.data


class SessionResultTest(TestCase):
    def setUp(self):
        super(SessionResultTest, self).setUp()
        self.results = [
            Result() for _ in range(10)
            ]
        # one result with both errors and failures
        try:
            1 / 0
        except:
            self.results[1].add_error()
            self.results[1].add_failure()
            # one result with failure
            self.results[2].add_failure()
            # one result with error
            self.results[3].add_error()
            self.results[5].add_error()

        # one result will skip
        self.results[4].add_skip("Reason")

        # and one result will skip with error
        self.results[5].add_skip("Reason")
        num_finished = 7

        for result in self.results[:num_finished]:
            result.mark_finished()
        self.result = SessionResults(Session())
        for index, r in enumerate(self.results):
            self.result._results_dict[index] = r

    def test_counts(self):
        self.assertEquals(self.result.get_num_results(), len(self.results))
        self.assertEquals(self.result.get_num_successful(), 2)
        # errors take precedence over failures
        self.assertEquals(self.result.get_num_errors(), 3)
        self.assertEquals(self.result.get_num_skipped(), 2)
        self.assertEquals(self.result.get_num_failures(), 1)



from .ctx import context
from .exception_handling import handling_exceptions
import logbook

_logger = logbook.Logger(__name__)


class _Cleanup(object):

    def __init__(self, func, args, kwargs, critical=False, success_only=False):
        assert not (success_only and critical)
        super(_Cleanup, self).__init__()
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.critical = critical
        self.success_only = success_only

    def __call__(self):
        return self.func(*self.args, **self.kwargs)  # pylint: disable=W0142

    def __repr__(self):
        return "{0} ({1},{2})".format(self.func, self.args, self.kwargs)


def add_cleanup(_func, *args, **kwargs):
    """
    Adds a cleanup function to the cleanup stack. Cleanups are executed in a LIFO order.

    Positional arguments and keywords are passed to the cleanup function when called.

    :param critical: If True, this cleanup will take place even when tests are interrupted by the user (Using Ctrl+C for instance)
    :param success_only: If True, execute this cleanup only if no errors are encountered
    :param args: positional arguments to pass to the cleanup function
    :param kwargs: keyword arguments to pass to the cleanup function
    """
    critical = kwargs.pop('critical', False)
    success_only = kwargs.pop('success_only', False)

    new_kwargs = kwargs.pop('kwargs', {}).copy()
    new_args = list(kwargs.pop('args', ()))
    if args or kwargs:
        _logger.warning(
            'Passing *args/**kwargs to slash.add_cleanup is deprecated')
        new_args.extend(args)
        new_kwargs.update(kwargs)

    added = _Cleanup(_func, new_args, new_kwargs, critical=critical, success_only=success_only)
    _get_cleanups().append(added)


def add_critical_cleanup(_func, *args, **kwargs):
    """
    Same as :func:`.add_cleanup`, only the cleanup will be called even on interrupted tests
    """
    return add_cleanup(_func, critical=True, *args, **kwargs)


def add_success_only_cleanup(_func, *args, **kwargs):
    """
    Same as :func:`.add_cleanup`, only the cleanup will be called only if the test succeeds
    """
    return add_cleanup(_func, success_only=True, *args, **kwargs)


def call_cleanups(critical_only=False, success_only=False):
    cleanups = _get_cleanups()
    while cleanups:
        cleanup = cleanups.pop()
        if critical_only and not cleanup.critical:
            continue
        if not success_only and cleanup.success_only:
            continue
        with handling_exceptions(swallow=True):
            _logger.debug("Calling cleanup: {0}", cleanup)
            cleanup()


def _get_cleanups():
    returned = getattr(context, "cleanups", None)
    if returned is None:
        returned = context.cleanups = []
    return returned



import itertools

from ...exceptions import UnknownFixtures, InvalidFixtureScope

from ...ctx import context
from .namespace import Namespace
from .parameters import get_parametrization_fixtures
from .utils import FixtureInfo
from .fixture_base import FixtureBase


_fixture_id = itertools.count()



class Fixture(FixtureBase):

    def __init__(self, store, fixture_func):
        super(Fixture, self).__init__()
        self.fixture_func = fixture_func
        self.info = self.fixture_func.__slash_fixture__
        self.scope = self.info.scope
        self.namespace = Namespace(store, store.get_current_namespace())

    def __repr__(self):
        return '<Function Fixture around {0}>'.format(self.fixture_func)

    def get_value(self, kwargs):
        return self.fixture_func(**kwargs)  # pylint: disable=star-args

    def _resolve(self, store):
        assert self.fixture_kwargs is None

        assert self.parametrization_ids is None
        self.parametrization_ids = []

        kwargs = {}
        parametrized = set()

        for parametrization_fixture in get_parametrization_fixtures(self.fixture_func):
            store.register_fixture_id(parametrization_fixture)
            parametrized.add(parametrization_fixture.name)
            self.parametrization_ids.append(parametrization_fixture.info.id)

        if 'this' in self.info.required_args:
            meta_fixture = ThisFixture(store, self)
            store.register_fixture_id(meta_fixture)
            self.namespace.add_name('this', meta_fixture.info.id)

        for name in self.info.required_args:
            if name in parametrized:
                continue
            try:
                needed_fixture = self.namespace.get_fixture_by_name(name)

                if needed_fixture.scope < self.scope:
                    raise InvalidFixtureScope('Fixture {0} is dependent on {1}, which has a smaller scope ({2} > {3})'.format(
                        self.info.name, name, self.scope, needed_fixture.scope))

                assert needed_fixture is not self
                kwargs[name] = needed_fixture.info.id
            except LookupError:
                raise UnknownFixtures(name)
        return kwargs


class ThisFixture(FixtureBase):

    def __init__(self, store, fixture):
        super(ThisFixture, self).__init__()
        assert context.session is None or store is context.session.fixture_store
        self.store = store
        self.info = FixtureInfo()
        self.fixture = fixture
        self.name = self.fixture.info.name
        self.scope = fixture.scope

    def _resolve(self, store):
        return {}

    def add_cleanup(self, func):
        self.store.add_cleanup(self.scope, func)

    def get_value(self, kwargs):
        assert not kwargs
        return self

    def __repr__(self):
        return '<This Fixture>'



from contextlib import contextmanager
from .utils.debug import debug_if_needed
from . import hooks as trigger_hook
from .ctx import context as slash_context
from .conf import config
from .exceptions import TestFailed, SkipTest
import functools
import logbook
try:
    import raven # pylint: disable=F0401
except ImportError:
    raven = None
import sys

_logger = logbook.Logger(__name__)

def update_current_result(exc_info):
    if slash_context.session is None:
        return
    if slash_context.test is not None:
        current_result = slash_context.session.results.get_result(slash_context.test)
    else:
        current_result = slash_context.session.results.global_result

    exc_class = exc_info[0]
    if issubclass(exc_class, (AssertionError, TestFailed)):
        current_result.add_failure()
    elif issubclass(exc_class, Exception) and not issubclass(exc_class, SkipTest):
        current_result.add_error()

def trigger_hooks_before_debugger(_):
    trigger_hook.exception_caught_before_debugger()  # pylint: disable=no-member
def trigger_hooks_after_debugger(_):
    trigger_hook.exception_caught_after_debugger()  # pylint: disable=no-member

_EXCEPTION_HANDLERS = [
    update_current_result,
    trigger_hooks_before_debugger,
    debug_if_needed,
    trigger_hooks_after_debugger,
    ]

@contextmanager
def handling_exceptions(**kwargs):
    swallow = kwargs.pop("swallow", False)
    try:
        yield
    except:
        handle_exception(sys.exc_info(), **kwargs)
        if not swallow:
            raise

def handle_exception(exc_info, context=None):
    """
    Call any handlers or debugging code before propagating an exception onwards.

    This makes sure that the exception can be handled as close as possible to its originating point.

    .. note:: this *DOES NOT* take care of adding the error to the session or test results!
    """
    already_handled = is_exception_handled(exc_info[1])
    msg = "Handling exception"
    if context is not None:
        msg += " (Context: {0})"
    if already_handled:
        msg += " (already handled)"
    _logger.debug(msg, context, exc_info=exc_info if not already_handled else None)

    if not already_handled:
        mark_exception_handled(exc_info[1])
        for handler in _EXCEPTION_HANDLERS:
            handler(exc_info)

def mark_exception_handled(e):
    mark_exception(e, "handled", True)

def is_exception_handled(e):
    """
    Checks if the exception ``e`` already passed through the exception handling logic
    """
    return bool(get_exception_mark(e, "handled", False))

_NO_DEFAULT = object()

def is_exception_marked(e, name):
    return get_exception_mark(e, name, _NO_DEFAULT) is not _NO_DEFAULT

def mark_exception(e, name, value):
    """
    Associates a mark with a given value to the exception ``e``
    """
    _ensure_exception_marks(e)[name] = value

def get_exception_mark(e, name, default=None):
    """
    Given an exception and a label name, get the value associated with that mark label.
    If the label does not exist on the specified exception, ``default`` is returned.
    """
    return _ensure_exception_marks(e).get(name, default)

def _ensure_exception_marks(e):
    returned = getattr(e, "__slash_exc_marks__", None)
    if returned is None:
        returned = e.__slash_exc_marks__ = {}
    return returned

@contextmanager
def get_exception_swallowing_context(report_to_sentry=True):
    """
    Returns a context under which all exceptions are swallowed (ignored)
    """
    try:
        yield
    except:
        if not get_exception_mark(sys.exc_info()[1], "swallow", True):
            raise
        if report_to_sentry:
            capture_sentry_exception()
        _logger.debug("Ignoring exception", exc_info=sys.exc_info())

def noswallow(exception):
    """
    Marks an exception to prevent swallowing by :func:`slash.exception_handling.get_exception_swallowing_context`,
    and returns it
    """
    mark_exception(exception, "swallow", False)
    return exception

def mark_exception_fatal(exception):
    """
    Causes this exception to halt the execution of the entire run.

    This is useful when detecting errors that need careful examination, thus preventing further tests from
    altering the test subject's state
    """
    mark_exception(exception, "fatal", True)
    return exception

def is_exception_fatal(exception):
    return bool(get_exception_mark(exception, "fatal", False))

def disable_exception_swallowing(func_or_exception):
    """
    Marks an exception to prevent swallowing. Can also be used as a decorator around a function to mark all escaped
    exceptions
    """
    if isinstance(func_or_exception, BaseException):
        return noswallow(func_or_exception)
    @functools.wraps(func_or_exception)
    def func(*args, **kwargs):
        try:
            return func_or_exception(*args, **kwargs)
        except BaseException as e:
            disable_exception_swallowing(e)
            raise
    return func

def capture_sentry_exception():
    client = get_sentry_client()
    if client is not None:
        client.captureException()

def get_sentry_client():
    if raven is not None and config.root.sentry.dsn:
        return raven.Client(config.root.sentry.dsn)
    return None
